(* :Package Version: 1.10.20 *)

(* :Copyright: Copyright 2002, Selwyn Hollis *)

(* :Name: DiffEqs`DETools` *)
 
(* :Author: Selwyn Hollis *)


BeginPackage["DiffEqs`DETools`"]


SeparableDSolve::usage = "SeparableDSolve[ LHS \[DifferentialD]x == RHS \[DifferentialD]y, options] 
 returns an implicit general solution in the form F[x,y]== C.";

FOLDSolve::usage = "FOLDSolve[equation, y[t], t, options]
solves a first-order linear (FOL) differential equation for the expression y[t].";
 
BernoulliDSolve::usage = "BernoulliDSolve[equation, y[t], t, options]
 solves a Bernoulli-type differential equation for the expression y[t].";

RiccatiDSolve::usage = "RiccatiDSolve[equation, y[t], t, knownsolution, options]
 solves a Riccati-type differential equation for the expression y[t].";
 
RicattiDSolve::usage = "Riccati has 1 c and 2 t's.";

SeriesDSolve::usage = "SeriesDSolve[eqn, y[t], {t, t0, n}]
returns a pair of independent series solutions of a linear 2nd-order
differential equation. The series expansions are about t=t0, with terms up to degree n.";


Options[SeparableDSolve] := {ConstantName -> C[1], InitialPoint -> None}; 
Options[FOLDSolve] := {ConstantName -> C[1]}; 
Options[BernoulliDSolve] := {ConstantName -> C[1]};
Options[RiccatiDSolve] := {ConstantName -> C[1]}; 
Options[SeriesDSolve] := {};


Begin["`Private`"]


SeparableDSolve[eqn_Equal, opt___Rule] :=
Module[{dvars, vars, funcs, there, cName, init, l1, r1}, 

   cName = ConstantName /. {opt} /. Options[SeparableDSolve];
   init = InitialPoint /. {opt} /. Options[SeparableDSolve];
   
   dvars = Cases[eqn, Global`DifferentialD[_], Infinity];
   dvars = Characters[ToString[Release[dvars]]]; 
   vars = ToExpression[Extract[dvars, {{16}, {34}}]];
   funcs = List @@ DeleteCases[l1*eqn[[1]]==r1*eqn[[2]], Global`DifferentialD[_], Infinity];
   funcs = funcs/.{l1->1,r1->1};
   there = Integrate[funcs[[1]], vars[[1]]] - Integrate[funcs[[2]], vars[[2]]];
   cName = If[init === None, cName, there //. {vars[[1]] -> init[[1]], vars[[2]] -> init[[2]]}]; 
   there == cName];
   
   

FOLDSolve[eqn:(_Equal|{_Equal}|{_Equal,_Equal}), y_[t_Symbol], t_Symbol, opt___Rule] := Module[
	{expr, t0, y0, p, q, P, eq, DE, j, jj, k, kk, \[Phi], \[Psi], cName, init}, 

   cName = ConstantName /. {opt} /. Options[FOLDSolve];
  
	FOLDSolve::"arg1"="Invalid argument(s).";
	If[Not[MemberQ[{1,2},Length[Position[eqn//Simplify, y'[t]]]]],
   		Message[FOLDSolve::"arg1"];Return[$Failed]];  
   
   eq = If[Head[eqn]===List && Length[eqn]===1, eqn[[1]], eqn];
   
   FOLDSolve::"init1"="Invalid initial condition.";
   init= If[
 		Head[eq]===Equal,
 			None, 
      		j=Position[eq, y'[_]][[1,1]]; jj=Mod[j,2]+1;
      		k=Position[eq[[jj]],y[_]][[1,1]]; kk=Mod[k,2]+1;
      		If[Sort[{j,jj,k,kk}]==={1,1,2,2},
      			{eq[[jj]][[k]] /. y[x_]->x, eq[[jj]][[kk]]},
      			Message[FOLDSolve::"init1"];Return[$Failed]]
      	];
      
   DE= If[Head[eq]===Equal, eq, eq[[Position[eq, y'[_]][[1,1]]]]];      

   t0 = If[init === None, 0, init[[1]]]; 
   y0 = If[init === None, cName, init[[2]]];
   expr = DE[[1]] - DE[[2]];
   expr = expr/Coefficient[expr, y'[t]];
   p = Coefficient[expr, y[t]]; 
   q = -expr /. {y[t] -> 0, y'[t] -> 0};
   \[Phi] = If[Head[s] === Symbol, Global`s, Global`\[Eta]]; 
   \[Psi] = If[Head[\[Tau]] === Symbol, Global`\[Tau], Global`\[Xi]];
   P = If[init === None,
   		E^Integrate[p, t], 
		E^Integrate[p /. t -> \[Psi], {\[Psi], t0, \[Phi]}]
		]//Simplify;
   q = If[init === None, q, q /. t -> \[Phi]]; 
   If[init === None,
   		(cName + Integrate[P*q, t])/P,
		(1/P /. \[Phi] -> t)*(y0 + Integrate[P*q, {\[Phi], t0, t}])]
  ];


BernoulliDSolve[eqn:(_Equal|{_Equal}|{_Equal,_Equal}), y_[t_Symbol], t_Symbol, opt___Rule] := 
  Module[{n, folineqn, eq, DE, usoln, t0, y0, u0, j, jj, k, kk, homog, evenDenominator, p, z, cName, init},

	cName = ConstantName /. {opt} /. Options[BernoulliDSolve]; 
   
	BernoulliDSolve::"arg1"="Invalid argument(s).";
	BernoulliDSolve::"arg2"="That's not a Bernoulli equation. Try again.";
    BernoulliDSolve::"init1"="Invalid initial-value problem.";
	BernoulliDSolve::"init2"="Illegal initial value.";
    BernoulliDSolve::"init3"="Real exponent requires a nonnegative initial value.";

	If[Not[MemberQ[{1,2},Length[Position[eqn//Simplify, y'[t]]]]],
   		Message[BernoulliDSolve::"arg1"]; Return[$Failed]];   
    
	eq = If[Head[eqn]===List && Length[eqn]===1, eqn[[1]], eqn];
   
 	 	init= If[
 		Head[eq]===Equal,
 			None, 
      		j=Position[eq, y'[_]][[1,1]]; jj=Mod[j,2]+1;
      		k=Position[eq[[jj]],y[_]][[1,1]]; kk=Mod[k,2]+1;
      		If[Sort[{j,jj,k,kk}]==={1,1,2,2},
      			{eq[[jj]][[k]] /._[x_]->x, eq[[jj]][[kk]]},
      			Message[BernoulliDSolve::"init1"];Return[$Failed]]
      	];
      
	DE= Which[Head[eq]===Equal, eq, Length[eq]==1, eq[[1]],
		Length[eq]===2, eq[[Position[eq, y'[_]][[1,1]]]]];

    n = Union[Cases[Expand /@ DE, y[t]^_, {1, Infinity}]];
    
    homog= Evaluate[DE/.{y[t]^_->0, y'[t]->0} /. y[t]->0] === True;
    
    If[Length[n] =!= 1  || Not[homog], Message[ BernoulliDSolve::"arg2"]; Return[$Failed], n = n[[1,2]]]; 
    evenDenominator = Denominator[n] == 2*Floor[Denominator[n]/2]; 
    t0 = If[init =!= None, init[[1]], {}];
    y0 = If[init =!= None, init[[2]], {}]; 
    If[init =!= None,
    	If[evenDenominator && y0 < 0, Message[BernoulliDSolve::"init2"];Return[$Failed]]; 
    	If[Head[n] === Real && y0 < 0,
    		Message[BernoulliDSolve::"init3"];Return[$Failed]];
    	If[y0 == 0 && Simplify[DE/.{y[t] -> 0, y'[t] -> 0}],Return[0]]
    ];
    folineqn = Apart /@ Simplify[Expand /@ (#1*(1-n)*u[t]^(n/(n-1))&)/@ 
          (DE //. {y[t] -> u[t]^(1/(1-n)), y'[t] -> u[t]^(n/(1-n))*
   			u'[t]/(1-n)}) /. {(u[t]^(1/(1-n)))^n -> u[t]^(n/(1-n))}, 
       		{u[t] \[Element] Reals, u[t] >= 0}];
    If[init =!= None, 
      u0 = If[NumericQ[y0], Which[y0 >= 0, y0^(1 - n), y0 < 0 && EvenQ[Numerator[1 - n]], 
         	Abs[y0]^(1-n), y0 < 0 && OddQ[Numerator[1-n]], -Abs[y0]^(1-n)], y0^(1 - n)];
      usoln = FOLDSolve[{folineqn, u[t0]==u0}, u[t], t];
      usoln^(1/(1-n))*If[NumericQ[y0]&& y0 < 0 && EvenQ[Numerator[1-n]],-1,1], 
      usoln=FOLDSolve[folineqn, u[t], t, ConstantName->cName];
      usoln^(1/(1-n))]];
      
      
RicattiDSolve[___]:= Print["Riccati has two cs and one t."];
      

RiccatiDSolve[eqn:(_Equal|{_Equal}|{_Equal,_Equal}), y_[t_Symbol], t_Symbol, knownsoln_, opt___Rule] := 
  Module[{beq, eq, DE, expr, p, f, check, cName, init}, 

	cName = ConstantName /. {opt} /. Options[RiccatiDSolve]; 
   
	RiccatiDSolve::"arg1"="Invalid argument(s).";
	RiccatiDSolve::"arg2"="That's not a Riccati equation. Try again.";
	RiccatiDSolve::"arg3"="Invalid particular solution";
    RiccatiDSolve::"init1"="Invalid initial-value problem.";
	RiccatiDSolve::"init2"="Illegal initial value.";
    RiccatiDSolve::"init3"="Real exponent requires a nonnegative initial value.";

   	If[Not[MemberQ[{1,2},Length[Position[eqn//Simplify, y'[t]]]]],
   		Message[RiccatiDSolve::"arg1"]; Return[$Failed]];
   
    eq = If[Head[eqn]===List && Length[eqn]===1, eqn[[1]], eqn];
   
 	init= If[
 		Head[eq]===Equal,
 			None, 
      		j=Position[eq, y'[_]][[1,1]]; jj=Mod[j,2]+1;
      		k=Position[eq[[jj]],y[_]][[1,1]]; kk=Mod[k,2]+1;
      		If[Sort[{j,jj,k,kk}]==={1,1,2,2},
      			{eq[[jj]][[k]] /._[x_]->x, eq[[jj]][[kk]]},
      			Message[RiccatiDSolve::"init1"]; Return[$Failed]]
      	];
      
	DE= If[Head[eq]===Equal, eq, eq[[Position[eq, y'[_]][[1,1]]]]];
      
   check = Union[Cases[Expand /@ DE, y[t]^_, {1, Infinity}]]; 
   If[Length[check] =!= 1 || N[check[[1,2]]] =!= 2., Message[RiccatiDSolve::"arg2"]; Return[$Failed]];
   If[Simplify[DE /. {y[t] -> knownsoln, y'[t] -> D[knownsoln, t]}] =!= True, 
     Message[RiccatiDSolve::"arg3"]; Return[$Failed]]; 
 
	expr = DE[[1]] - DE[[2]];
 	expr = expr/Coefficient[expr, Derivative[1][y][t]]; 
	p = Coefficient[expr, y[t]];
 	f = -Coefficient[expr, y[t]^2] - Coefficient[expr, y[t]^2.]; 
	beq = z'[t] + (p - 2*knownsoln*f)*z[t] - f*z[t]^2 == 0;
    If[init === None, knownsoln + BernoulliDSolve[beq, z[t], t, ConstantName -> cName], 
     knownsoln + BernoulliDSolve[{beq, z[init[[1]]]==init[[2]] - (knownsoln /. t -> init[[1]])}, z[t], t]
     ]
   ];
   


SeriesDSolve[eqn:(_Equal|{_Equal}|{_Equal,_Equal}|{_Equal,_Equal,_Equal}),
							y_[t_Symbol], {t_Symbol, t0_, nterms_Integer}] := 
  Module[{eq, diffeq, order,j, jj, fty, ftyv, yy, vv, yy0, vv0, ics, soln, zero, sys, expr, usoln,
          vsoln, instances, chars, initpt, degenerate, homogeneous},
	
   SeriesDSolve::"arg1"="Invalid argument(s).";
   SeriesDSolve::"arg3"="First argument must be either an equation or a list of equations.";
   SeriesDSolve::"arg2"="A second-order DE requires either 2 initial conditions or none.";
   SeriesDSolve::"eqs1"="SeriesDSolve only works on 1st and 2nd-order equations.";
   SeriesDSolve::"degen1"="SeriesDSolve does't solve degenerate nonhomogeneous equations.";
   SeriesDSolve::"inits1"="Bad initial values";
   SeriesDSolve::"inits2"="Initial values can't be used.";
   
   
   	If[
   	Not[MemberQ[{1,2},Length[Position[eqn//Simplify, y''[_]]]]] &&
   	Not[MemberQ[{1,2},Length[Position[eqn//Simplify, y'[_]]]]],
   		Message[SeriesDSolve::"arg1"]; Return[$Failed]];
   		
	order=If[Not[MemberQ[{1,2},Length[Position[eqn//Simplify,y''[_]]]]], 1, 2];
	
	If[order===1,
 	If[Head[eqn]===List && Length[eqn]===2,
	
		fty=(y'[t]/.First@Solve[eqn,y'[t]])/.y[t]->yy;
	 	yy0= y[t0]/.First@Solve[eqn,y[t0]];
		Return[TaylorSoln1[fty,{t,t0,nterms},{yy,yy0}]],
	
		fty=(y'[t]/.First@Solve[eqn,y'[t]])/.y[t]->yy;
	 	yy0= y[t0]/.First@Solve[eqn,y[t0]];
		Return[TaylorSoln1[fty,{t,t0,nterms},{yy,y[t0]}]]
		]];
		
	(* From here on, order = 2. *)
	
	If[Not[MemberQ[{1,3}, Length[Flatten[{eqn}]]]],
   		Message[SeriesDSolve::"arg2"]; Return[$Failed]];
	  
    eq = If[Head[eqn]===List && Length[eqn]===1, eqn[[1]], eqn]; 
   
 	initpt= If[Head[eq]===Equal,
 					None,
    			ics=Delete[eq, Position[eq, y''[t]][[1,1]]];
    			j=Position[ics, y'[t0]][[1,1]];
    			jj=Position[ics, y[t0]][[1,1]];
    			If[Union[{j,jj}]==={1,2},
    				{ics[[jj]]/.y[t0]==x_->x, ics[[j]]/.y'[t0]==x_->x},
    				Message[SeriesDSolve::"inits1"]; Return[$Failed]]
    	      ];
      
	diffeq = If[Head[eq]===Equal, eq, eq[[Position[eq, y''[t]][[1,1]]]]];

    instances = Union[Extract[diffeq, Position[diffeq, Derivative[_], Infinity]]]; 
    chars = Characters[ToString[instances]];
    
   Which[
     Length[instances] > 2, 
     	Message[SeriesDSolve::"eqs1"]; Return[$Failed], 
     Length[instances] == 1 && Max[ToExpression[Extract[chars, {{13}}]]] > 2, 
     	Message[SeriesDSolve::"eqs1"]; Return[$Failed], 
     Length[instances] == 2 && Max[ToExpression[Extract[chars, {{13}, {28}}]]] > 2, 
     	Message[SeriesDSolve::"eqs1"]; Return[$Failed],
     True, {}];

 degenerate = N[Coefficient[diffeq[[1]] - diffeq[[2]], y''[t]] /. t -> t0]===0.;     
  homogeneous = N[(diffeq[[1]]-diffeq[[2]])//.{y[t]->0, y'[t]->0, y''[t]->0}]===0.;
 		   
  If[degenerate && Not[homogeneous],
   Message[SeriesDSolve::"degen1"]; Return[$Failed]];
   
  If[degenerate && homogeneous && initpt =!= None,
  		Message[SeriesDSolve::"inits2"]]; 
  		
  If[Not[degenerate],
  		Which[
  		Length[eqn]===3,
  			ftyv=(y''[t]/.First@Solve[eqn,y''[t]])/.{y[t]->yy, y'[t]->vv};
  			yy0= y[t0]/.First@Solve[eqn,y[t0]]; vv0= y'[t0]/.Solve[eqn,y'[t0]][[1]];
			Return[TaylorSoln2[ftyv,{t,t0,nterms},{yy,yy0},{vv,vv0}]],
		homogeneous,
			ftyv=(y''[t]/.First@Solve[eqn,y''[t]])/.{y[t]->yy, y'[t]->vv};
			Return[{TaylorSoln2[ftyv,{t,t0,nterms},{yy,1},{vv,0}],
			TaylorSoln2[ftyv,{t,t0,nterms},{yy,0},{vv,1}]}],
		Not[homogeneous],
			ftyv=(y''[t]/.First@Solve[eqn,y''[t]])/.{y[t]->yy, y'[t]->vv};
			Return[TaylorSoln2[ftyv,{t,t0,nterms},{yy,0},{vv,0}]]
		]
	];
	
    soln := a[0]*x^r + Sum[a[n]*x^(n+r), {n,1,nterms}];
	expr = diffeq[[1]] - diffeq[[2]] /. t -> x+t0;
    zero = 
	  Normal[Series[Coefficient[expr, y''[x+t0]], {x,0,nterms}]]*y''[x+t0] + 
      Normal[Series[Coefficient[expr, y'[x+t0]], {x,0,nterms}]]*y'[x+t0] + 
      Normal[Series[Coefficient[expr, y[x+t0]], {x,0,nterms}]]*y[x+t0] + 
      Normal[Series[expr/.{y[x+t0]->0, y'[x+t0]->0, y''[x+t0]->0}, {x,0,nterms}]];
      
	zero = zero /. {y[x+t0] -> soln, y'[x+t0] -> D[soln, x], y''[x+t0] -> D[soln,x,x]};

	zero = Expand[zero*x^(2-r)]; 
	While[(zero /. x -> 0) === 0, zero = Expand[zero/x]];
  			sys = Flatten[Append[{(zero /. x -> 0) == 0, a[0] == 1},
					Table[Coefficient[zero, x^n] == 0, {n, 1, nterms}]]];
	soln = soln /. Solve[sys, Flatten[{r, Table[a[n], {n, 0, nterms}]}]]; 
	Union[soln/.{a[_]->0, x->t-t0}]
];


TaylorSoln1[f_, {t_, t0_, n_Integer}, {y_, y0_}] := 
  Module[{derivs}, 
   derivs = Join[{y[t]}, Table[
        D[f /. y-> y[t], {t, k - 1}], 
        {k, 1, n}]] //. {t-> t0, y[t0]-> y0}; 
    Do[yk = D[y[t], {t, k}] /. t -> t0; 
      derivs = derivs //. yk-> derivs[[k + 1]], {k, 1, n}]; 
    Sum[(derivs[[j + 1]]/j!)*(t-t0)^j, {j, 0, n}]];

TaylorSoln2[f_, {t_, t0_, n_Integer}, {y_, y0_},{v_, v0_}] := 
  Module[{derivs}, 
   derivs = Join[{y[t],y'[t]}, Table[
        D[f/.{y-> y[t],v-> y'[t]}, {t, k - 1}], 
        {k, 1, n}]] //. {t-> t0, y[t0]-> y0, y'[t0]-> v0}; 
    Do[yk = D[y[t], {t, k}] /. t-> t0; 
      derivs = derivs //. yk-> derivs[[k + 1]], {k, 1, n}]; 
    Sum[(derivs[[j + 1]]/j!)*(t-t0)^j, {j, 0, n}]];

End[]

EndPackage[]


